import socket
from struct import pack
from json import dumps

ST_ROKU = "roku:ecp"
ST_DIAL = "urn:dial-multiscreen-org:service:dial:1"
__DEFAULT_TIMEOUT__ = 1


class SimpleServiceDiscoveryProtocol:

    TIMEOUT = __DEFAULT_TIMEOUT__

    def __init__(self, st):
        self.st = st
        self.timeout = SimpleServiceDiscoveryProtocol.TIMEOUT
        self.message = DiscoveryMessage(self.st)

    def broadcast(self) -> list:
        ttl = pack("b", 1)
        results = []

        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP) as s:
            s.settimeout(self.timeout)
            s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)
            s.sendto(self.message.encode("utf8"), self.message.DISCOVER_GROUP)

            while True:
                try:
                    data, _ = s.recvfrom(4096)
                    resp = HTTPResponse(data)
                    results.append(resp)
                except socket.timeout:
                    break
        return results

    def __repr__(self):
        return f"{self.__class__.__name__}({self.st!r})"

    @classmethod
    def settimeout(cls, timeout):
        if not isinstance(timeout, (int, float)):
            raise ValueError(
                f"timeout must be of type int or float and not {type(timeout)}."
            )
        cls.TIMEOUT = timeout


class DiscoveryMessage(str):

    DISCOVER_GROUP = ("239.255.255.250", 1900)

    def __new__(cls, st):
        MESSAGE_CONFIG = {
            "HOST": "%s:%s" % cls.DISCOVER_GROUP,
            "ST": st,
            "MX": SimpleServiceDiscoveryProtocol.TIMEOUT,
            "MAN": "ssdp:discover",
        }

        DISCOVER_MESSAGE = (
            "M-SEARCH * HTTP/1.1\r\n"
            + "\r\n".join(
                [
                    "%s:%s" % (key, value)
                    for (key, value) in MESSAGE_CONFIG.items()
                    if value
                ]
            )
            + "\r\n"
        )

        return str.__new__(cls, DISCOVER_MESSAGE)


class HTTPResponse(dict):
    def __init__(self, response_text):

        response = response_text.decode("utf-8").split("\r\n")
        status_line = response[0]

        self.http_version, self.status_code, self.status = status_line.split()
        self.headers = {}

        for line in response[1:]:
            line = line.split()
            if len(line) == 2:
                header_name = line[0][:-1].replace("-", "_").lower()
                header_value = line[1].lower()
                self.headers[header_name] = header_value

    @property
    def json(self):
        return dumps(self.__dict__)

    def __repr__(self):
        return self.json



import requests
from urllib.parse import urlparse


class Controller:
    def __init__(self, address):
        parsed_url = urlparse(address)
        self.address = f"http://{parsed_url.netloc}"

    def __getitem__(self, prop):
        return self.__getattribute__(prop)

    def __enter__(self):
        self._session = requests.Session()
        return self

    def __exit__(self, *args):
        self._session.close()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.address!r})"

    @property
    def request(self) -> requests:
        if "_session" in dir(self):
            return self._session
        return requests



from typing import List, Callable
from re import sub
from urllib.parse import unquote, urlencode

from requests import Response

from pylaunch.core import Controller
from pylaunch.ssdp import SimpleServiceDiscoveryProtocol, ST_DIAL
from pylaunch.xmlparse import XMLFile


class AppNotFoundError(Exception):
    pass


def discover(timeout: int = 3) -> List["Dial"]:
    """
    Scans network for dial compatible devices and returns a list.
    """
    results = []
    SimpleServiceDiscoveryProtocol.settimeout(timeout)
    ssdp = SimpleServiceDiscoveryProtocol(ST_DIAL)
    response = ssdp.broadcast()
    for resp in response:
        location = resp.headers.get("location")
        if not location:
            continue
        results.append(Dial(location))
    return results


class Dial(Controller):
    def __init__(self, address: str):
        self.bind(address)
        self.instance_url = None
        self.refresh_url = None

    def __str__(self):
        return f"{self.friendly_name} ({self.address})"

    def bind(self, address: str) -> None:
        """
        A function called on __init__ to bind to a specific device.
        """
        resp = self.request.get(address)
        self.address = resp.headers.get("Application-URL")
        xml = XMLFile(resp.text)
        key = (
            lambda element: "_".join(sub("([a-z])([A-Z])", r"\1 \2", element).split())
            .lower()
            .replace("-", "_")
        )
        value = lambda element: xml.find(element).text
        tag_name = lambda element: element.tag.replace(xml.namespace, "")

        for element in [tag_name(el) for el in xml.find("device")]:
            try:
                k = key(element)
                v = value(element)
            except TypeError as e:
                print(e)
            finally:
                setattr(self, k, v) if v != "\n" else None

    def _build_app_url(self, app_name: str = None) -> str:
        """
        Simple helper function to build app urls.
        """
        return "/".join([self.address, app_name])

    def launch_app(
        self, app_name: str, callback: Callable[[None], Response] = None, **kwargs
    ) -> None:
        """
        Launches specified application.
        """
        url = self._build_app_url(app_name)
        data = unquote(urlencode(kwargs))
        headers = (
            {"Content-Type": "text/plain; charset=utf-8"}
            if kwargs
            else {"Content-Length": "0"}
        )
        resp = self.request.post(url, data=data, headers=headers)

        if resp.status_code < 300:
            self.instance_url = resp.headers.get("location")
            self.refresh_url = unquote(resp.text)
            callback(resp) if callback else None
        elif resp.status_code == 404:
            raise AppNotFoundError(f"No application found with name {app_name}")
        else:
            resp.raise_for_status()

    def kill_app(
        self, app_name: str = None, callback: Callable[[None], Response] = None
    ) -> None:
        """
        This will kill any active application tracked by this 
        instance if one exists and will return True if successful 
        otherwise it will return False.
        """
        if app_name:
            app_url = self._build_app_url(app_name) + "/run"
        elif not app_name and not self.instance_url:
            raise Exception("There is no instance found to kill.")
        else:
            app_url = self.instance_url
        resp = self.request.delete(app_url)
        if resp.status_code in [200, 204]:
            self.instance_url = None
            self.refresh_url = None
        elif resp.status_code == 404:
            raise Exception(f"There is no running {app_name} instance.")
        callback(resp) if callback else None

    def get_app_status(self, app_name: str) -> dict:
        """
        Makes a request to the DIAL device with a application name parameter and returns
        a dictionary including the supported DIAL version, app name and state.
        State examples: running, stopped or installable
        """
        url = self._build_app_url(app_name)
        resp = self.request.get(url, headers={"Content-Type": "text/plain"})
        if resp.status_code == 404:
            raise AppNotFoundError(f"No application found with name {app_name}")
        xml = XMLFile(resp.text)
        return {
            "version": xml.find("service").attrib.get("dialVer"),
            "name": xml.find("name").text,
            "state": xml.find("state").text,
        }

    def refresh_instance(self, inplace: bool = False) -> str:
        """
        Makes a request using the refresh_url stored in the instance
        of this class.
        """
        if not self.refresh_url:
            raise Exception("No refresh_url found in the dial instance.")
        resp = self.request.post(self.refresh_url)
        instance_url = resp.headers.get("location")
        if inplace:
            self.instance_url = instance_url
        else:
            return instance_url



from __future__ import annotations
from typing import Callable, List, Dict

from pylaunch.core import Controller
from pylaunch.ssdp import ST_ROKU, SimpleServiceDiscoveryProtocol
from pylaunch.xmlparse import XMLFile, normalize


class Application:
    def __init__(
        self, name: str, id: str, type: str, subtype: str, version: str, roku: Roku
    ):
        self.name = name
        self.id = id
        self.type = type
        self.subtype = subtype
        self.version = version
        self.roku = roku

    def __repr__(self):
        return "{cn}(name='{name}', id='{id}', type='{type}', subtype='{subtype}', version='{version}')".format(
            cn=self.__class__.__name__,
            name=self.name,
            id=self.id,
            type=self.type,
            subtype=self.subtype,
            version=self.version,
        )

    def __getattribute__(self, name):
        return super().__getattribute__(name)

    @property
    def icon(self) -> Dict[str, str]:
        request_url = f"{self.roku.address}/query/icon/{self.id}"
        response = self.roku.request.get(request_url, stream=True)
        if str(response.headers["Content-Length"]) != "0":
            filetype = response.headers["Content-Type"].split("/")[-1]
            return {"content": response.content, "filetype": filetype}
        return {"content": "", "filetype": ""}

    def launch(self, callback: Callable[[None], dict] = None, **kwargs) -> None:
        request_url = f"{self.roku.address}/launch/{self.id}"
        response = self.request.post(
            request_url, params=kwargs, headers={"Content-Length": "0"}
        )
        if callback:
            results = {"request_url": request_url, "status_code": response.status_code}
            callback(results)


class Roku(Controller):
    def __getitem__(self, key):
        if key in self.info:
            return self.info.get(key)
        elif key in self.apps:
            return self.apps.get(key)
        else:
            raise AttributeError(key)

    @classmethod
    def discover(cls, timeout: int = 3) -> List[Roku]:
        """
        Scans the network for roku devices.
        """
        results = []
        SimpleServiceDiscoveryProtocol.settimeout(timeout)
        ssdp = SimpleServiceDiscoveryProtocol(ST_ROKU)
        response = ssdp.broadcast()
        for resp in response:
            location = resp.headers.get("location")
            if not location:
                continue
            results.append(cls(location))
        return results

    @property
    def active_app(self) -> Application:
        request_url = f"{self.address}/query/active-app"
        response = self.request.get(request_url)
        xml = XMLFile(response.text)
        element = xml.find("app")
        return Application(
            name=element.text,
            id=element.attrib.get("id"),
            type=element.attrib.get("type"),
            subtype=element.attrib.get("subtype"),
            version=element.attrib.get("version"),
            roku=self,
        )

    @property
    def apps(self) -> Dict[str, Application]:
        applications = {}
        request_url = f"{self.address}/query/apps"
        response = self.request.get(request_url)
        xml = XMLFile(response.text)
        for element in xml.find("apps"):
            app = Application(
                name=element.text,
                id=element.attrib.get("id"),
                type=element.attrib.get("type"),
                subtype=element.attrib.get("subtype"),
                version=element.attrib.get("version"),
                roku=self,
            )
            applications[app.name] = app
        return applications

    @property
    def info(self) -> Dict[str, [str, bool]]:
        device_info = {}
        request_url = f"{self.address}/query/device-info"
        response = self.request.get(request_url)
        xml = XMLFile(response.text)
        for element in xml.find("device-info"):
            key, value = normalize(xml, element)
            device_info[key] = value
        return device_info

    def install_app(self, id: str, **kwargs) -> None:
        request_url = f"{self.address}/install/{str(id)}"
        response = self.request.post(
            request_url, params=kwargs, headers={"Content-Length": "0"}
        )

    def key_press(self, key: str, callback: Callable[[None], dict] = None) -> None:
        request_url = f"{self.address}/keypress/{str(key)}"
        response = self.request.post(request_url)
        if callback:
            results = {"request_url": request_url, "status_code": response.status_code}
            callback(results)

    def power(self) -> None:
        power_modes = {"PowerOn": "Headless", "Headless": "PowerOn"}
        toggle_power_mode = (
            lambda x: setattr(self, "power_mode", power_modes[self.power_mode])
            if x["status_code"] == 200
            else None
        )

        self.key_press("power", toggle_power_mode)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pylaunch",
    version="0.1.61",
    package_dir={"": "src"},
    packages=["pylaunch", "pylaunch.roku", "pylaunch.dial"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Sandersland/pylaunch",
    author="Steffen Andersland",
    author_email="stefandersland@gmail.com",
    license="MIT",
    keywords=["dial", "roku"],
    install_requires=["requests"],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows :: Windows 10",
    ],
    extras_require={
        "dev": ["nose==1.3.7", "twine==3.1.1", "pre-commit==2.0.1", "black==19.10b0",]
    },
)
